#include "stdafx.h"
#include <MersenneTwister.h>
#include <HMM.h>
#include "TaskSimulate.h"

bool MicroSimulate( bool * pbCancelled, CModel & oModel, int iChannelCount, int iSimPoints, double dSampling, 
					double * adData, short * aiIdlState, CIdlList * pIdlList, bool bRemoveDwells );
bool MacroSimulate( bool * pbCancelled, CModel & oModel, int iChannelCount, int iSimPoints, double dSampling, 
					double * adData, short * aiIdlState, bool bDeterministic, bool bRemoveDwells );

//----------------------------------------------------------------------
void COptionsSimulate :: ToBuffer( BYTE * buf ) { 
	sprintf( (char *)buf, "2 %6d %8d %1d %6d %4d %8d %c", 
			min(m_iSegments,999999), min(m_iPointCount,99999999), 
			m_bRandomSeed ? 1 : 0, min(m_iRandomSeed,999999), 
			min( m_iChannelCount, 9999), min( m_iSampleRate,99999999), 
			m_bRemoveDwells ? 'Y' : 'N' );
	// Note - The following are not included : 
	// CItemArray m_aStimulation;		( save with model ? )
	// CString m_cFileName;				( Save with model ? )
	}

void COptionsSimulate :: FromBuffer( BYTE * buf ) { 
	int iVer, iRandomSeed;
	char chRemoveDwells;
	sscanf( (char *)buf, "%d %6d %8d %1d %6d %4d %8d %c", 
			&iVer, &m_iSegments, &m_iPointCount, &iRandomSeed, 
			&m_iRandomSeed, &m_iChannelCount, &m_iSampleRate, &chRemoveDwells );
	m_bRemoveDwells = ( chRemoveDwells != 'N' );
	m_bRandomSeed = (iRandomSeed!=0);
	}

#ifdef WIN32
bool COptionsSimulate :: RegistryLoad() { 
	CString c = RegistryLoadString( "OPTIONSSIMULATE" ); 
	if( c.Length()==0 ) 
		return false;
	FromBuffer( (BYTE *)c.ConstBuffer() );
	return true;
	}

bool COptionsSimulate :: RegistrySave() { 
	BYTE buf[BUF_OPTSIM_SIZE+64];
	ToBuffer( buf );
	return RegistrySaveString( "OPTIONSSIMULATE", (char *)buf );
	}
#endif
	
//-----------------------------------------------------------------
CTaskSimulate :: CTaskSimulate() {
	m_pOptionsSimulate=NULL;
	}
	
CTaskSimulate :: ~CTaskSimulate() {
	if( m_pOptionsSimulate!=NULL ) 
		delete m_pOptionsSimulate;
	}
			
//-----------------------------------------------------------------
DWORD WINAPI TaskSimulate( LPVOID pParam ) { 
	CTaskSimulate * pTask = (CTaskSimulate *) pParam;

	if( ! pTask->m_pOptionsSimulate->m_bRandomSeed ) { 
		int iSeed = pTask->m_pOptionsSimulate->m_iRandomSeed;
		if( iSeed==0 ) 
			iSeed=0x3d41;
		init_genrand( iSeed );
		}
	
	int iSegs=pTask->m_pOptionsSimulate->m_iSegments;
	int iPts=pTask->m_pOptionsSimulate->m_iPointCount;
	int iClass, iClasses=pTask->m_pOptionsSimulate->m_oModel.m_aClasses.Count();
	
	int iUsedClass=0, iUsedClasses = pTask->m_pOptionsSimulate->m_oModel.UsedClassCount();
	int * aiUsedClasses = new int[iUsedClasses];
	for( iClass=0; iClass<iClasses && iUsedClass<iUsedClasses; ++iClass ) 
		if( pTask->m_pOptionsSimulate->m_oModel.UsesClass(iClass) ) 
			aiUsedClasses[ iUsedClass++ ] = iClass;

	for( int iSeg=0; iSeg<iSegs; ++iSeg ) { 
		//-- Create buffer to fill with data 
		double * pData = new double[iPts];		// deleted by receiving owner
		
		//-- Create / initialize new idealization to be filled
		CIdlList * pIdlList = new CIdlList;
		pIdlList->m_iLength=iPts;
		pIdlList->AllocateClasses( iUsedClasses );
		for( iUsedClass=0; iUsedClass<iUsedClasses; ++iUsedClass ) {
			iClass = aiUsedClasses[iUsedClass];
			pIdlList->IdlClass(iUsedClass)->m_dAmp=pTask->m_pOptionsSimulate->m_oModel.Class(iClass)->m_dAmp;
			pIdlList->IdlClass(iUsedClass)->m_dStdDev=pTask->m_pOptionsSimulate->m_oModel.Class(iClass)->m_dStdDev;
			}
		
		//-- Sim some data 
		MicroSimulate( &pTask->m_bCancelled, 
						pTask->m_pOptionsSimulate->m_oModel, 
						pTask->m_pOptionsSimulate->m_iChannelCount, 
						iPts, 
						1.0/pTask->m_pOptionsSimulate->m_iSampleRate, 
						pData, NULL, pIdlList, 
						pTask->m_pOptionsSimulate->m_bRemoveDwells );
		
		//-- check if cancelled 
		if( pTask->m_bCancelled ) 
			break;
			
		//-- send to data window 
		PostMessage( pTask->m_hwndDataView, WM_DATA_ADDIDLSEG, (WPARAM)pIdlList, (LPARAM)pData );

		//-- Update the pct complete 
		PostMessage( pTask->m_hDlg, WM_TASKSETPCT, MulDiv(iSeg,100,iSegs), 0 );
		}
	
	PostMessage( pTask->m_hwndDataView, WM_DATA_CHANGED, 1, 0 );	

	PostMessage( (HWND)pTask->m_hDlg, WM_TASKCOMPLETE, pTask->m_bCancelled?1:0, 0 );
	
	delete [] aiUsedClasses;
	return -1;
	}	

//============================================================================
// Code for threaded simulation task.

  /*
  Consider : 
	int aiTransition[n][n] : Record count of transitions from i to j 
	double adTransition[n][n] : Record duration of event following transition 
  Consider option for scaled integer output.
  Todo - add an idealization output structure
  */


//=============================================================================
//-- Transition from state iState of Q to a new randomly selected state weighted by rates.
int GetNewState( double ** Q, int iState, int iStates ) { 
	//-- iRand is random within [0,sum(Rates)] to provide weighted random selection
	double dRand = RandomLT( -Q[iState][iState] );
		
	//-- determine which state dRand corresponds to.  
	//-- i.e., for rates={0,100,200,10} then any value between 100 and 299 is the 3rd state.
	double dSum=0.0;
	int j;
	for( j=0; j<iStates; ++j ) 
		if( j != iState ) 
			if( (dSum+=Q[iState][j]) > dRand ) 
				return j;
				
	//-- if there are no exit rates from this state or Q is mangled or some other problem, return something.
	ThreadErrorMessage( "GetNewState failed.  Lowest will be used.\n" );	//todo - internationalize this 
	for( j=0; j<iStates; ++j ) 
		if( j != iState && Q[iState][j]!=0 )
			return j;
			
	return iState;
	}

//=============================================================================
// Simulate (iSimPoints) of data from (oModel) with (iChannelCount) channels.
// adData[iSimPoints] will be filled.  
// If aiIdlState != NULL it will be filled with state for each point.
bool MicroSimulate( bool * pbCancelled, CModel & oModel, int iChannelCount, int iSimPoints, double dSampling, 
					double * adData, short * aiIdlState, CIdlList * pIdlList, bool bRemoveDwells ) { 
					
	double dAmp, dStd;					
	int iClass, iState, iStates = oModel.m_aStates.Count();					
	TNT::Array2D<double> aQ = oModel.Q( NULL, 0.0 );
	CModelClass * pClass;
	
	//-- Create array of starting probabilities 
	double * adP0 = new double[iStates];
	double d1, d2;
	oModel.NormalizeStartProbabilities( adP0 );

	//-- For each channel  
	for( int iCh=0; iCh<iChannelCount; ++iCh ) { 
		int iPointsAccum=0, iTotalPoints;
        double dDurationEvent=0.0, dDurationAccum=0.0;
        bool bFirst=true;

		//---- until buffer is full 
		while( iPointsAccum<iSimPoints ) { 
			//-- create next event state 
			iState = bFirst ? RandomWeighted(iStates,adP0) : GetNewState(aQ,iState,iStates);
			bFirst = false;

			//-- Create next event 
			double dRnd = RandomExponential();
			double dLifeTime = double(-aQ[iState][iState]);
			dDurationEvent = dRnd / dLifeTime;
			iClass = oModel.State( iState )->m_iClass;
			pClass = oModel.Class(iClass);
			dAmp=pClass->m_dAmp;
			dStd=pClass->m_dStdDev;
                
			//-- Calc how many samples will have passed before this event completes 
			dDurationAccum+=dDurationEvent;
			iTotalPoints = min( int( dDurationAccum/dSampling ), iSimPoints );
			
			//-- Add a dwell to the idealization.  
			if( ! bRemoveDwells || iPointsAccum<iTotalPoints )		// whether dwells of 0 points (missed events) should appear.
			if( iChannelCount==1 && iPointsAccum<iSimPoints )
				pIdlList->AddDwell( iClass, iTotalPoints-iPointsAccum, iState );
				
			//-- add samples until iTotalPoints is reached (possibly none!)
			// Note : add event, duration to list ( always or only if it crosses a sample ? )
			while( iPointsAccum < iTotalPoints ) { 
				//-- generate 2 (gaussian) random values
				RandomGauss2( dAmp, dStd, &d1, &d2 );
				//-- use the first random
				if( iCh==0 ) 
					adData[iPointsAccum] = d1; 
				else 
					adData[iPointsAccum] += d1; 
				if( iCh==0 && aiIdlState!=NULL ) 
					aiIdlState[iPointsAccum] = iState;
				++iPointsAccum;
				//-- use the 2nd random if it fits.				
				if( iPointsAccum < iTotalPoints ) {
					if( iCh==0 ) 
						adData[iPointsAccum] = d2; 
					else 
						adData[iPointsAccum] += d2; 
					if( iCh==0 && aiIdlState!=NULL ) 
						aiIdlState[iPointsAccum] = iState;
					++iPointsAccum;
					}
				}
			
			//-- Update progress display
			//ThreadProgress( (iCh+iPointsAccum/iTotalPoints)*100/iChannelCount );
			
			//-- Check for cancel
			if( *pbCancelled ) 
				break;
			}
		}
		
	delete [] adP0;
	return true;
	}

//=============================================================================
// TODO - what is the difference between microscopic multi-channel and stochastic macroscopic ? 
bool MacroSimulate( bool * pbCancelled, CModel & oModel, int iChannelCount, int iSimPoints, double dSampling, 
					double * adData, short * aiIdlState, bool bDeterministic, bool bRemoveDwells ) { 

	//--
	int i, iPt, iChangePt=-1;
	int iState, iStates = oModel.m_aStates.Count();
    int iProb=0;					// which probability vector (0/1) is current.
    
	//-- fill in array of pointers for quick ref.
	CModelState ** aModelState = new CModelState * [iStates];
	CModelClass ** aModelClass = new CModelClass * [iStates];
	for( iState=0; iState<iStates; ++iState ) { 
		aModelState[iState] = oModel.State( iState );
		aModelClass[iState] = oModel.Class( aModelState[iState]->m_iClass );
		}
	double dBaseAmp = aModelClass[0]->m_dAmp;
	double dBaseStd = aModelClass[0]->m_dStdDev;

    //-- 2 probabilities vectors - alternate between to reduce temporary copies
    double * adProb[2];
    adProb[0] = new double[iStates];
    adProb[1] = new double[iStates];

	//-- fill in base amplitude 
	if( bDeterministic )						//-- deterministic or stochastic response 
		for( iPt=0; iPt<iSimPoints; ++iPt ) 
			adData[iPt] = dBaseAmp;
	else 
		for( iPt=0; iPt<iSimPoints; ++iPt ) 
			adData[iPt] = RandomGauss( dBaseAmp, dBaseStd );

	//-- Go - simulate data 
	HMM::CModelMath<double> o(iStates);
    int iProgress=0;
	for( iPt=0; iPt<iSimPoints; ++iPt ) {
		
		//-- stimulus changed - update model and recalculate math for 'A' matrix
		if( iPt >= iChangePt ) { 
			//-- Update o.m_aQ from oModel+Stimulus, then o.CalcA()
			o.m_aQ = oModel.Q( NULL, dSampling*iPt );
			o.CalcA( dSampling );
			//--set iChangePt to next change in stimulus
			// todo - calculate this from stimulus expressions.
			iChangePt = iSimPoints;
			}
	
        //-- Advance the probability vector
		iProb=1-iProb;						// switch to other storage vector
	    if ( iPt==0 )						// P = P0
		    oModel.NormalizeStartProbabilities( adProb[iProb] );
		else {								// P = P * A 
			double dSumProb=0.0;			// sum of all probabilities
			for( iState=0; iState<iStates; ++iState ) {
				//-- New prob. of iState is sum of each current * chance of transition to iState
				adProb[iProb][iState]=0.0;
				for( i=0; i<iStates; ++i )		
					adProb[iProb][iState] += adProb[1-iProb][i] * o.m_aA[i][iState];
				dSumProb += adProb[iProb][iState];
				}

			//--Rescale probabilities
			if( dSumProb==0.0 )				// normally only for pt 0.
				for( iState=0; iState<iStates; ++iState )
					adProb[iProb][iState] = 1.0/iStates;
			else if( dSumProb!=1.0 ) 
				for( iState=0; iState<iStates; ++iState )
					adProb[iProb][iState] /= dSumProb;	
			}
				
        //-- Set amplitude
        double dAmp=0.0, dStd;
		for( iState=0; iState<iStates; ++iState ) { 
			if( aModelState[iState]->m_iClass != 0 ) { 
				if ( bDeterministic ) 
					dAmp += aModelClass[iState]->m_dAmp * adProb[iProb][iState] * iChannelCount;
				else { 
					dStd = aModelClass[iState]->m_dStdDev;
					if( dStd<=dBaseStd ) 
						dStd=0.0;
					else 
						dStd = sqrt( (dStd*dStd-dBaseStd*dBaseStd) * adProb[iProb][iState] * dBaseStd );
					dAmp += RandomGauss( dBaseAmp+(aModelClass[iState]->m_dAmp-dBaseAmp)*adProb[iProb][iState]*iChannelCount, dStd );
					}
				}
			}
		adData[iPt] += dAmp;

		//-- Update progress display
		//if( iPt*100/iSimPoints!=iProgress ) 
		//	ThreadProgress( iProgress=iPt*100/iSimPoints );	// ineff. recalc, but occurs generally <=100 times.
		}

	delete [] aModelState;
	delete [] aModelClass;
	delete [] adProb[0];
	delete [] adProb[1];
	return true;
	}

